@@ -23,6 +23,7 @@
 #include <dbus/dbus-glib.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <locale.h>
 
 #include <nm-utils.h>
 #include <nm-setting-connection.h>
@@ -48,13 +49,13 @@ item_count_func (const char *key, const char *value, gpointer user_data)
 }
 
 static void
-test_items (const char *detail, NMSettingVPN *s_vpn, Item *test_items, gboolean secrets)
+test_items (const char *detail, NMSettingVPN *s_vpn, Item *items, gboolean secrets)
 {
 	Item *iter;
 	guint32 expected_count = 0, actual_count = 0;
 	const char *value;
 
-	for (iter = test_items; iter->name; iter++) {
+	for (iter = items; iter->name; iter++) {
 		if (secrets)
 			value = nm_setting_vpn_get_secret (s_vpn, iter->name);
 		else
@@ -141,7 +142,6 @@ test_basic_import (NMVpnPluginUiInterface *plugin, const char *dir)
 	const char *expected_route1_gw = "0.0.0.0";
 	const char *expected_route2_dest = "172.16.0.0";
 	const char *expected_route2_gw = "0.0.0.0";
-	const char *value;
 
 	connection = get_basic_connection ("basic-import", plugin, dir, "basic.pcf");
 	ASSERT (connection != NULL, "basic-import", "failed to import connection");
@@ -159,7 +159,7 @@ test_basic_import (NMVpnPluginUiInterface *plugin, const char *dir)
 
 	/* IP4 setting */
 	s_ip4 = (NMSettingIP4Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP4_CONFIG);
-	ASSERT (s_con != NULL,
+	ASSERT (s_ip4 != NULL,
 	        "basic-import", "missing 'ip4-config' setting");
 
 	ASSERT (nm_setting_ip4_config_get_num_addresses (s_ip4) == 0,
@@ -260,7 +260,7 @@ remove_secrets (NMConnection *connection)
 	g_slist_free (keys);
 }
 
-#define EXPORTED_NAME "basic-export-test.pcf"
+#define BASIC_EXPORTED_NAME "basic-export-test.pcf"
 static void
 test_basic_export (NMVpnPluginUiInterface *plugin, const char *dir)
 {
@@ -274,7 +274,7 @@ test_basic_export (NMVpnPluginUiInterface *plugin, const char *dir)
 	connection = get_basic_connection ("basic-export", plugin, dir, "basic.pcf");
 	ASSERT (connection != NULL, "basic-export", "failed to import connection");
 
-	path = g_build_path ("/", dir, EXPORTED_NAME, NULL);
+	path = g_build_path ("/", dir, BASIC_EXPORTED_NAME, NULL);
 	success = nm_vpn_plugin_ui_interface_export (plugin, path, connection, &error);
 	if (!success) {
 		if (!error)
@@ -284,7 +284,7 @@ test_basic_export (NMVpnPluginUiInterface *plugin, const char *dir)
 	}
 
 	/* Now re-import it and compare the connections to ensure they are the same */
-	reimported = get_basic_connection ("basic-export", plugin, dir, EXPORTED_NAME);
+	reimported = get_basic_connection ("basic-export", plugin, dir, BASIC_EXPORTED_NAME);
 	ret = unlink (path);
 	ASSERT (connection != NULL, "basic-export", "failed to re-import connection");
 
@@ -301,6 +301,52 @@ test_basic_export (NMVpnPluginUiInterface *plugin, const char *dir)
 	g_free (path);
 }
 
+#define NAT_EXPORTED_NAME "nat-export-test.pcf"
+static void
+test_nat_export (NMVpnPluginUiInterface *plugin, const char *dir, const char *nat_mode)
+{
+	NMConnection *connection;
+	NMSettingVPN *s_vpn;
+	NMConnection *reimported;
+	char *path;
+	gboolean success;
+	GError *error = NULL;
+	int ret;
+
+	connection = get_basic_connection ("nat-export", plugin, dir, "basic.pcf");
+	ASSERT (connection != NULL, "nat-export", "failed to import connection");
+
+	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	ASSERT (s_vpn != NULL, "nat-export", "imported connection had no VPN setting");
+
+	nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_NAT_TRAVERSAL_MODE, nat_mode);
+
+	path = g_build_path ("/", dir, NAT_EXPORTED_NAME, NULL);
+	success = nm_vpn_plugin_ui_interface_export (plugin, path, connection, &error);
+	if (!success) {
+		if (!error)
+			FAIL ("nat-export", "export failed with missing error");
+		else
+			FAIL ("nat-export", "export failed: %s", error->message);
+	}
+
+	/* Now re-import it and compare the connections to ensure they are the same */
+	reimported = get_basic_connection ("nat-export", plugin, dir, NAT_EXPORTED_NAME);
+	ret = unlink (path);
+	ASSERT (connection != NULL, "nat-export", "failed to re-import connection");
+
+	/* Clear secrets first, since they don't get exported, and thus would
+	 * make the connection comparison below fail.
+	 */
+	remove_secrets (connection);
+
+	ASSERT (nm_connection_compare (connection, reimported, NM_SETTING_COMPARE_FLAG_EXACT) == TRUE,
+	        "nat-export", "original and reimported connection differ");
+
+	g_object_unref (reimported);
+	g_object_unref (connection);
+	g_free (path);
+}
 
 static void
 test_everything_via_vpn (NMVpnPluginUiInterface *plugin, const char *dir)
@@ -311,7 +357,6 @@ test_everything_via_vpn (NMVpnPluginUiInterface *plugin, const char *dir)
 	GError *error = NULL;
 	char *pcf;
 	const char *expected_id = "All your traffic are belong to VPN";
-	const char *value;
 
 	pcf = g_build_path ("/", dir, "everything-via-vpn.pcf", NULL);
 	ASSERT (pcf != NULL,
@@ -333,7 +378,7 @@ test_everything_via_vpn (NMVpnPluginUiInterface *plugin, const char *dir)
 
 	/* IP4 setting */
 	s_ip4 = (NMSettingIP4Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP4_CONFIG);
-	ASSERT (s_con != NULL,
+	ASSERT (s_ip4 != NULL,
 	        "everything-via-vpn", "missing 'ip4-config' setting");
 
 	ASSERT (nm_setting_ip4_config_get_never_default (s_ip4) == FALSE,
@@ -388,6 +433,135 @@ test_no_natt (NMVpnPluginUiInterface *plugin, const char *dir)
 	g_free (pcf);
 }
 
+static void
+test_nat_cisco (NMVpnPluginUiInterface *plugin, const char *dir)
+{
+	NMConnection *connection;
+	NMSettingConnection *s_con;
+	NMSettingVPN *s_vpn;
+	GError *error = NULL;
+	char *pcf;
+	const char *expected_id = "NAT-Cisco";
+	const char *value;
+
+	pcf = g_build_path ("/", dir, "nat-cisco.pcf", NULL);
+	ASSERT (pcf != NULL,
+	        "nat-cisco", "failed to create pcf path");
+
+	connection = nm_vpn_plugin_ui_interface_import (plugin, pcf, &error);
+	if (error)
+		FAIL ("nat-cisco", "error importing %s: %s", pcf, error->message);
+	ASSERT (connection != NULL,
+	        "nat-cisco", "error importing %s: (unknown)", pcf);
+
+	/* Connection setting */
+	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
+	ASSERT (s_con != NULL,
+	        "nat-cisco", "missing 'connection' setting");
+
+	ASSERT (strcmp (nm_setting_connection_get_id (s_con), expected_id) == 0,
+	        "nat-cisco", "unexpected connection ID");
+
+	/* VPN setting */
+	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	ASSERT (s_vpn != NULL,
+	        "nat-cisco", "missing 'vpn' setting");
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+	ASSERT (value != NULL,
+	        "nat-cisco", "unexpected missing value for item %s", NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+	ASSERT (strcmp (value, NM_VPNC_NATT_MODE_CISCO) == 0,
+	        "nat-cisco", "unexpected value for item %s", NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+
+	g_free (pcf);
+}
+
+static void
+test_nat_natt (NMVpnPluginUiInterface *plugin, const char *dir)
+{
+	NMConnection *connection;
+	NMSettingConnection *s_con;
+	NMSettingVPN *s_vpn;
+	GError *error = NULL;
+	char *pcf;
+	const char *expected_id = "NAT-T";
+	const char *value;
+
+	pcf = g_build_path ("/", dir, "natt.pcf", NULL);
+	ASSERT (pcf != NULL,
+	        "natt", "failed to create pcf path");
+
+	connection = nm_vpn_plugin_ui_interface_import (plugin, pcf, &error);
+	if (error)
+		FAIL ("natt", "error importing %s: %s", pcf, error->message);
+	ASSERT (connection != NULL,
+	        "natt", "error importing %s: (unknown)", pcf);
+
+	/* Connection setting */
+	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
+	ASSERT (s_con != NULL,
+	        "natt", "missing 'connection' setting");
+
+	ASSERT (strcmp (nm_setting_connection_get_id (s_con), expected_id) == 0,
+	        "natt", "unexpected connection ID");
+
+	/* VPN setting */
+	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	ASSERT (s_vpn != NULL,
+	        "natt", "missing 'vpn' setting");
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+	ASSERT (value != NULL,
+	        "natt", "unexpected missing value for item %s", NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+	ASSERT (strcmp (value, NM_VPNC_NATT_MODE_NATT) == 0,
+	        "natt", "unexpected value for item %s", NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+
+	g_free (pcf);
+}
+
+static void
+test_nat_force_natt (NMVpnPluginUiInterface *plugin, const char *dir)
+{
+	NMConnection *connection;
+	NMSettingConnection *s_con;
+	NMSettingVPN *s_vpn;
+	GError *error = NULL;
+	char *pcf;
+	const char *expected_id = "Force NAT-T";
+	const char *value;
+
+	pcf = g_build_path ("/", dir, "force-natt.pcf", NULL);
+	ASSERT (pcf != NULL,
+	        "force-natt", "failed to create pcf path");
+
+	connection = nm_vpn_plugin_ui_interface_import (plugin, pcf, &error);
+	if (error)
+		FAIL ("force-natt", "error importing %s: %s", pcf, error->message);
+	ASSERT (connection != NULL,
+	        "force-natt", "error importing %s: (unknown)", pcf);
+
+	/* Connection setting */
+	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
+	ASSERT (s_con != NULL,
+	        "force-natt", "missing 'connection' setting");
+
+	ASSERT (strcmp (nm_setting_connection_get_id (s_con), expected_id) == 0,
+	        "force-natt", "unexpected connection ID");
+
+	/* VPN setting */
+	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	ASSERT (s_vpn != NULL,
+	        "force-natt", "missing 'vpn' setting");
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+	ASSERT (value != NULL,
+	        "force-natt", "unexpected missing value for item %s", NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+	ASSERT (strcmp (value, NM_VPNC_NATT_MODE_NATT_ALWAYS) == 0,
+	        "force-natt", "unexpected value for item %s", NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
+
+	g_free (pcf);
+}
+
 static void
 test_always_ask (NMVpnPluginUiInterface *plugin, const char *dir)
 {
@@ -429,6 +603,42 @@ test_always_ask (NMVpnPluginUiInterface *plugin, const char *dir)
 	g_free (pcf);
 }
 
+static void
+test_non_utf8_import (NMVpnPluginUiInterface *plugin, const char *dir)
+{
+	NMConnection *connection;
+	NMSettingConnection *s_con;
+	NMSettingVPN *s_vpn;
+	const char *expected_id = "Att äta en ko";
+	const char *charset = NULL;
+
+	/* Change charset to ISO-8859-15 to match iso885915.pcf */
+	g_get_charset (&charset);
+	setlocale (LC_ALL, "de_DE@euro");
+	connection = get_basic_connection ("non-utf8-import", plugin, dir, "iso885915.pcf");
+	setlocale (LC_ALL, charset);
+
+	ASSERT (connection != NULL, "non-utf8-import", "failed to import connection");
+
+	/* Connection setting */
+	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
+	ASSERT (s_con != NULL,
+	        "non-utf8-import", "missing 'connection' setting");
+
+	ASSERT (strcmp (nm_setting_connection_get_id (s_con), expected_id) == 0,
+	        "non-utf8-import", "unexpected connection ID");
+
+	ASSERT (nm_setting_connection_get_uuid (s_con) == NULL,
+	        "non-utf8-import", "unexpected valid UUID");
+
+	/* VPN setting */
+	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	ASSERT (s_vpn != NULL,
+	        "non-utf8-import", "missing 'vpn' setting");
+
+	g_object_unref (connection);
+}
+
 int main (int argc, char **argv)
 {
 	GError *error = NULL;
@@ -455,9 +665,16 @@ int main (int argc, char **argv)
 	test_basic_import (plugin, argv[1]);
 	test_everything_via_vpn (plugin, argv[1]);
 	test_no_natt (plugin, argv[1]);
+	test_nat_cisco (plugin, argv[1]);
+	test_nat_natt (plugin, argv[1]);
+	test_nat_force_natt (plugin, argv[1]);
 	test_always_ask (plugin, argv[1]);
+	test_non_utf8_import (plugin, argv[1]);
 
 	test_basic_export (plugin, argv[1]);
+	test_nat_export (plugin, argv[1], NM_VPNC_NATT_MODE_CISCO);
+	test_nat_export (plugin, argv[1], NM_VPNC_NATT_MODE_NATT);
+	test_nat_export (plugin, argv[1], NM_VPNC_NATT_MODE_NATT_ALWAYS);
 
 	g_object_unref (plugin);
 
